/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * backward computation for math operations
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-18
 */

#include "../tensor/XTensor.h"

#ifndef __XBACKWARDMATH_H__
#define __XBACKWARDMATH_H__

namespace nts{

/* this class computes the gradient for math operations given a node */
class XMathGrad
{
public:
    /* compute dE/dx of a node */
    static
    void MakeGrad(XTensor * node);

    /* indicates whether the node is for a math operation */
    static
    bool IsMathOP(XTensor * node);

private:
    /* gradient for sum: c =  a + b * \beta */
    static
    void GradSum(XTensor * node);

    /* gradient for multiply (dot production): c =  a * b */
    static
    void GradMultiply(XTensor * node);

    /* gradient for matrix multiply: c = matmul(a, b) */
    static
    void GradMatrixMul(XTensor * node);
};

}

#endif